from copy import copy
from copy import deepcopy

keys = ('note','dur','amp','onset')

class note_transform:
    def __init__(self):
        self.shift={}
        self.scale={}
        for key in keys:
            self.shift[key]=0.
        for key in keys:
            self.scale[key]=1.
    def copy(other):
        a=note_transform()
        a.shift=copy(other.shift)
        a.scale=copy(other.scale)
        return a
    def add_shift(self,key,a):
        self.shift[key]+=a
    def add_scale(self,key,a):
        self.scale[key]*=a    
    def transform_note(self,note_info):
        new_note=copy(note_info)
        for key in keys:
            new_note[key]+=self.shift[key]
            new_note[key]*=self.scale[key]
        return new_note
    def join(self,other):
        for key in keys:
            self.scale[key]*=other.scale[key]
            self.shift[key]+=other.shift[key]
    def __repr__(self):
        s=""
        for key in keys:
            if self.shift[key] <> 0:
                s+="(+,%s,%f)"%(key,self.shift[key])
            if self.scale[key] <> 1:
                s+="(*,%s,%f)"%(key,self.scale[key])
        return s

from array import array

def int_add(root,a):
    b=a.pop(0)
    if b=="0":
        int_add(root.children[0],a)
    else:
        int_add(root.children[1],a)

def add(root,s):
    a=array('c',s)
    int_add(root,a)
    
class node:
    def __init__(self,joining=True,name="no_name"):
        self.children=[None,None]
        self.joining=joining
        self.parent=None
        self.transform=note_transform()
        self.side=-1
        self.leaf=False
        self.name=name
    def copy(self):
        n=node(self.joining)
        n.transform=self.transform.copy()
        s="%s"%(self.name)
        n.name=s
        return n
    
    def add_child(self,c,ind,name="no_name"):
        self.children[ind]=c
        c.parent=self
        c.side=ind
        c.name=name
        return c
    def score(self,args):
        tr=args[0]
        tr.join(self.transform)
    def id(self):
        p=self
        ids=[]
        while p<>None:
            id=""
            if p.side < 0:
                id+="|"
            else:
                if p.side==0:
                    id+="l"
                else:
                    id+="r"
            if hasattr(p,"joining"):
                if p.joining:
                    id+="S"
                else:
                    id+="U"
            if hasattr(p,"transform"):
                id+=str(p.transform)
            ids.append(id)
            p=p.parent
        s=""
        ids.reverse()
        for id in ids:
            s+=id
        return s
    def print_id(self):
        if self.leaf:
            print(self.id(),self.name)
   

class note_node(node):
    def __init__(self,note,dur,am):
        self.children=[None,None]
        self.info={'note':note,'dur':dur,'amp':am,'onset':0.0}
        self.leaf=True
        self.name=""
    def copy(self):
        n=note_node(self.info['note'],self.info['dur'],self.info['amp'])
        s="%s"%(self.name)
        n.name=s
        return n
    
    def score(self,args):
        transform=args[0]
        notes=args[1]
        debug=args[2]
        new_note=transform.transform_note(self.info)
        if self.side==1:
            grand_parent=self.parent.parent
            if grand_parent<>None and grand_parent.joining:
                transform.shift['onset'] += new_note['dur']
        if self.side==0 and self.parent.joining:
            transform.shift['onset'] += new_note['dur']
        s="%f,%d %s"%(new_note['onset'],new_note['note'],self.name)
        if debug:
            print(s)
        notes.append(new_note)
    


def traverse_tree(root,method_name,args=None):
    if root==None:
        return
    if hasattr(root,method_name):
        if args==None:
            getattr(root,method_name)()
        else:
            getattr(root,method_name)(args)    
    traverse_tree(root.children[0],method_name,args)
    traverse_tree(root.children[1],method_name,args)
    
def copy_subtree(stop_child,parent,copy_parent=None):
    if copy_parent==None:
        copy_parent=parent.copy()
        
    for ind in range(2):
        child=parent.children[ind]
        if child==None:
            continue
        copy_child=child.copy()
        copy_parent.add_child(copy_child,ind,copy_child.name)
        if child==stop_child:
            continue
        copy_subtree(stop_child,child,copy_child)
    return copy_parent
        
class recursion_operator:
    def __init__(self,parent,child,d,transform):
        self.parent=parent
        self.child=child
        self.depth=d
        self.transform=transform
    def process(self):
        if self.depth==0:
            return False 
        self.depth-=1
        #s="copy %s  to %s transforming %s"%(self.parent.name,self.child.name,str(self.transform))
        copy=copy_subtree(self.child,self.parent)
        copy.transform=self.transform
        self.child.parent.add_child(copy,self.child.side)
        return True        

def score(root,debug=False):
    notes=[]
    traverse_tree(root,"score",[note_transform(),notes,debug])
    return notes


    
def make_sonic_score(notes):
    import sframe
    t0=0
    sframe.make_item(0,t0,'versuch')
    it=sframe.get_note_item(0,t0)
    sonic_notes=[]
    ind=0
    it.clear()
    for n in notes:
        it.make_item(n['onset'],n['note'],n['amp'],n['dur']);
        ind+=1
    
               
def test3():
    root=node(True,"root")
    root.add_child(note_node(30,1,1),0,"nn0")
    n1=root.add_child(node(True),1,"S")
    n1.add_child(note_node(33,1,1),1,"nn1")
    n=n1.add_child(node(False),0,"S")
    c1=n.add_child(note_node(37,1,1),0,"nn2")
    c=n.add_child(note_node(36,1,1),1,"nn3")
    score(root)
    tr=note_transform()
    tr.add_shift("note",1)
    tr.add_scale("dur",.5)
    recursions=[]
    r=recursion_operator(root,c,0,tr)
    recursions.append(recursion_operator(root,c,3,tr))
    tr=note_transform()
    tr.add_shift("note",4)
    tr.add_scale("amp",.9)
    tr.add_scale("dur",1.25)
    recursions.append(recursion_operator(n1,c1,3,tr))
    good=True
    while good:
        good=False
        for r in recursions:
            if r.process():
                good=True
        
    print("processed ")
    notes=score(root,True)
    make_sonic_score(notes)
    #traverse_tree(root,"print_id")

def test2():
    root=node(True,"root")
    root.transform=note_transform()
    root.transform.add_shift("note",1)
    root.add_child(note_node(30,1,1),0,"nn0")
    n=root.add_child(node(True),1,"S")
    n.add_child(note_node(33,1,1),1,"nn1")
    traverse_tree(root,"score",[note_transform()])

       
def scale(scale):
    root=node(True,"root")
    s1=root.add_child(node(False,"hallo"),0)
    s=root.add_child(node(True),1)
    s3=s.add_child(node(True),0)
    s1.add_child(note_node(30,.125,1),0)
    s1.add_child(note_node(42,.125*0.5,.5),1)
    tr=note_transform()
    tr.add_shift("note",1)
    tr.add_scale("dur",scale)
    recursions=[]
    recursions.append(recursion_operator(root,s3,12,tr))
    good=True
    while good:
        good=False
        for r in recursions:
            if r.process():
                good=True
    return root
    #traverse_tree(root,"print_id")

def test():
    sc=scale(1.1)
    sc2=scale(1.05)
    root=node(True,"root")
    root.add_child(sc,0)
    root.add_child(sc,1)
    notes=score(root,True)
    make_sonic_score(notes)

print(80*"_")


test()    
        